#include <omp.h>
#include <string>
#include <chrono>
#include <random>
#include <iostream>
#include <iostream>
#include <algorithm>

void PrintHead() {std::cout << "RESULT\t";}
void PrintTail() {std::cout << std::endl;}
void Print(auto key, auto val) {std::cout << key << "=" << val << "\t";}
void PrintConfig(const auto& lambda_gen_name, const auto& input_size, const auto& it) {
    PrintHead();
    Print("input-type", lambda_gen_name);
    Print("n", input_size);
    Print("it", it);
}

void ExecBenchmark(size_t thread_cnt, auto& v) {
#if defined(_OPENMP)
        omp_set_num_threads(thread_cnt);
#endif
        std::chrono::steady_clock::time_point start = std::chrono::steady_clock::now();
        std::sort(v.begin(), v.end());
        std::chrono::steady_clock::time_point end = std::chrono::steady_clock::now();
        auto usec = std::chrono::duration_cast<std::chrono::nanoseconds>(end - start).count();
        Print("threads", thread_cnt);
        Print("wall-time", usec);
                
}

int main(int argc, char *argv[])
{
    size_t min_input_size = atoi(argv[1]);
    size_t max_input_size = atoi(argv[2]);
    size_t it_cnt = atoi(argv[3]);
    size_t num_threads = 1;

#if defined(_OPENMP)
    if (argc == 5) {
        num_threads = atoi(argv[4]);
    }
    else num_threads = omp_get_max_threads();
#endif

    std::random_device rd;
    std::mt19937 gen(rd());
    long min = std::numeric_limits<long>::min();
    long max = std::numeric_limits<long>::max();
    std::uniform_int_distribution<long> dis(min, max);

    for (size_t input_size = min_input_size;
         input_size < max_input_size;
         input_size *= 2) {

        std::vector<long> v(input_size);
        for (size_t it = 0; it != it_cnt; ++it) {
            for (size_t thread_cnt = 1; thread_cnt <= num_threads; thread_cnt*=2) {
                PrintConfig("uniform", input_size, it);
                std::generate (v.begin(), v.end(), [&dis, &gen](){return dis(gen);});
                ExecBenchmark(thread_cnt,
                              v);
                PrintTail();

                PrintConfig("zero", input_size, it);
                std::generate (v.begin(), v.end(), [](){return 0;});
                ExecBenchmark(thread_cnt,
                              v);
                PrintTail();
            }            
        }
    }

        
    return 0;
}
